from typing import Optional

import requests
from lxml import etree
from xml.sax.saxutils import escape

# import defusedxml.ElementTree as etree


class Office365:
    """
    Class to authenticate Office  365 Sharepoint
    """

    def __init__(self, share_point_site, username, password):
        # type: (str, str, str) -> None
        self.username = username
        self.password = password
        self.share_point_site = share_point_site

    def get_security_token(self, username, password):
        # type: (str, str) -> Optional[str]
        """
        Grabs a security Token to authenticate to Office 365 services
        """
        url = "https://login.microsoftonline.com/extSTS.srf"
        body = """
                <s:Envelope xmlns:s="http://www.w3.org/2003/05/soap-envelope"
                  xmlns:a="http://www.w3.org/2005/08/addressing"
                  xmlns:u="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd">
              <s:Header>
                <a:Action s:mustUnderstand="1">http://schemas.xmlsoap.org/ws/2005/02/trust/RST/Issue</a:Action>
                <a:ReplyTo>
                  <a:Address>http://www.w3.org/2005/08/addressing/anonymous</a:Address>
                </a:ReplyTo>
                <a:To s:mustUnderstand="1">https://login.microsoftonline.com/extSTS.srf</a:To>
                <o:Security s:mustUnderstand="1"
                   xmlns:o="http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd">
                  <o:UsernameToken>
                    <o:Username>%s</o:Username>
                    <o:Password>%s</o:Password>
                  </o:UsernameToken>
                </o:Security>
              </s:Header>
              <s:Body>
                <t:RequestSecurityToken xmlns:t="http://schemas.xmlsoap.org/ws/2005/02/trust">
                  <wsp:AppliesTo xmlns:wsp="http://schemas.xmlsoap.org/ws/2004/09/policy">
                    <a:EndpointReference>
                      <a:Address>%s</a:Address>
                    </a:EndpointReference>
                  </wsp:AppliesTo>
                  <t:KeyType>http://schemas.xmlsoap.org/ws/2005/05/identity/NoProofKey</t:KeyType>
                  <t:RequestType>http://schemas.xmlsoap.org/ws/2005/02/trust/Issue</t:RequestType>
                  <t:TokenType>urn:oasis:names:tc:SAML:1.0:assertion</t:TokenType>
                </t:RequestSecurityToken>
              </s:Body>
            </s:Envelope>""" % (
            escape(username),
            escape(password),
            self.share_point_site,
        )
        # headers = {"accept": "application/json;odata=verbose"}

        # response = requests.post(url, body, headers=headers)
        response = requests.post(url, body)

        xmldoc = etree.fromstring(response.content)

        token = xmldoc.find(
            ".//{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}BinarySecurityToken"
        )
        if token is not None:
            return token.text
        else:
          message = xmldoc.findall('.//{http://schemas.microsoft.com/Passport/SoapServices/SOAPFault}text')
          if len(message) < 1:
            raise Exception('Error authenticating against Office 365. Was not able to find an error code. Here is '
                    'the SOAP response from Office 365', response.content)
          raise Exception('Error authenticating against Office 365. Error from Office 365:', message[0].text)

    def get_cookies(self):
        # type: () -> requests.cookies.RequestsCookieJar
        """
        Grabs the cookies form your Office Sharepoint site
        and uses it as Authentication for the rest of the calls
        """
        sectoken = self.get_security_token(self.username, self.password)
        url = self.share_point_site + "/_forms/default.aspx?wa=wsignin1.0"
        response = requests.post(url, data=sectoken)
        return response.cookies

    # Legacy API
    GetSecurityToken = get_security_token
    GetCookies = get_cookies



from typing import Dict
from typing import List
from typing import Optional

from lxml import etree

# TODO: Port to defusedxml to satisfy Bandit
# import defusedxml.ElementTree as etree


class Soap:
    """A simple class for building SOAP Requests"""

    def __init__(self, command):  # type: (str) -> None
        self.envelope = None
        self.command = command
        self.request = None
        self.updates = None
        self.batch = None

        # HEADER GLOBALS
        SOAPENV_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/"
        SOAPENV = "{%s}" % SOAPENV_NAMESPACE
        ns0_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/"
        ns1_NAMESPACE = "http://schemas.microsoft.com/sharepoint/soap/"
        xsi_NAMESPACE = "http://www.w3.org/2001/XMLSchema-instance"
        NSMAP = {"SOAP-ENV": SOAPENV_NAMESPACE, "ns0": ns0_NAMESPACE, "ns1": ns1_NAMESPACE, "xsi": xsi_NAMESPACE}

        # Create Header
        self.envelope = etree.Element(SOAPENV + "Envelope", nsmap=NSMAP)
        HEADER = etree.SubElement(self.envelope, "{http://schemas.xmlsoap.org/soap/envelope/}Body")

        # Create Command
        self.command = etree.SubElement(HEADER, "{http://schemas.microsoft.com/sharepoint/soap/}" + command)

        self.start_str = b"""<?xml version="1.0" encoding="utf-8"?>"""

    def add_parameter(self, parameter, value=None):
        # type: (str, Optional[str]) -> None
        sub = etree.SubElement(self.command, "{http://schemas.microsoft.com/sharepoint/soap/}" + parameter)
        if value:
            sub.text = value

    # UpdateListItems Method
    def add_actions(self, data, kind):
        # type: (List[Dict[str, str]], str) -> None
        if not self.updates:
            updates = etree.SubElement(self.command, "{http://schemas.microsoft.com/sharepoint/soap/}updates")
            self.batch = etree.SubElement(updates, "Batch")
            self.batch.set("OnError", "Return")
            self.batch.set("ListVersion", "1")

        if kind == "Delete":
            for index, _id in enumerate(data, 1):
                method = etree.SubElement(self.batch, "Method")
                method.set("ID", str(index))
                method.set("Cmd", kind)
                field = etree.SubElement(method, "Field")
                field.set("Name", "ID")
                field.text = str(_id)

        else:
            for index, row in enumerate(data, 1):
                method = etree.SubElement(self.batch, "Method")
                method.set("ID", str(index))
                method.set("Cmd", kind)
                for key, value in row.items():
                    field = etree.SubElement(method, "Field")
                    field.set("Name", key)
                    field.text = str(value)

    # GetListFields Method
    def add_view_fields(self, fields):
        # type: (List[str]) -> None
        viewFields = etree.SubElement(self.command, "{http://schemas.microsoft.com/sharepoint/soap/}viewFields")
        viewFields.set("ViewFieldsOnly", "true")
        ViewFields = etree.SubElement(viewFields, "ViewFields")
        for field in fields:
            view_field = etree.SubElement(ViewFields, "FieldRef")
            view_field.set("Name", field)

    # GetListItems Method
    def add_query(self, pyquery):
        # type: (Dict) -> None
        query = etree.SubElement(self.command, "{http://schemas.microsoft.com/sharepoint/soap/}query")
        Query = etree.SubElement(query, "Query")
        if "OrderBy" in pyquery:
            order = etree.SubElement(Query, "OrderBy")
            for field in pyquery["OrderBy"]:
                fieldref = etree.SubElement(order, "FieldRef")
                if type(field) == tuple:
                    fieldref.set("Name", field[0])
                    if field[1] == "DESCENDING":
                        fieldref.set("Ascending", "FALSE")
                else:
                    fieldref.set("Name", field)

        if "GroupBy" in pyquery:
            order = etree.SubElement(Query, "GroupBy")
            for field in pyquery["GroupBy"]:
                fieldref = etree.SubElement(order, "FieldRef")
                fieldref.set("Name", field)

        if "Where" in pyquery:
            Query.append(pyquery["Where"])

    def __repr__(self):  # type: () -> str
        return (self.start_str + etree.tostring(self.envelope)).decode("utf-8")

    def __str__(self, pretty_print=False):  # type: (bool) -> str
        return (self.start_str + etree.tostring(self.envelope, pretty_print=True)).decode("utf-8")



import requests
import json

class _Folder():
    def __init__(self, session, folder_name, url):
        self._session = session
        self.folder_name = folder_name
        self.site_url = url
        self.timeout = 3
        
        self.info = self._create_folder()
        
    @property
    def contextinfo(self):
        response = self._session.post(self.site_url + "/_api/contextinfo")
        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))

        data = json.loads(response.text)
        return data
        
    def _create_folder(self):
        update_data = {}
        update_data['__metadata'] = {'type': 'SP.Folder'}
        update_data['ServerRelativeUrl'] = self.folder_name
        body = json.dumps(update_data)
        
        url = self.site_url + f"/_api/web/folders"
        
        headers = {'Accept': 'application/json;odata=verbose',
                   'Content-Type': 'application/json;odata=verbose',
                   'X-RequestDigest': self.contextinfo['FormDigestValue']}
        
        response = self._session.post(url=url,
                                      headers=headers,
                                      data=body,
                                      timeout=self.timeout)

        if response.status_code != 201:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))
        
        data = json.loads(response.text)
        return data
    
    def delete_folder(self, relative_url):
        if relative_url == self.folder_name:
            url = self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self.folder_name}')"

            headers = {'Accept': 'application/json;odata=verbose',
                       'If-Match': '*',
                       'X-HTTP-Method': 'DELETE',
                       'Content-Type': 'application/json;odata=verbose',
                       'X-RequestDigest': self.contextinfo['FormDigestValue']}

            response = self._session.post(url=url,
                                          headers=headers)

            if response.status_code != 200:
                response.raise_for_status()
                raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))
        else:
            print('You must pass the relative folder url to delete a folder')
        
        return None

    def delete_file(self, file_name):
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{file_name}')"

        headers = {'Accept': 'application/json;odata=verbose',
                    'If-Match': '*',
                    'X-HTTP-Method': 'DELETE',
                    'Content-Type': 'application/json;odata=verbose',
                    'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = self._session.post(url=url,
                                        headers=headers)

        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))

        return None

    @property
    def items(self):        
        response = self._session.get(self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self.folder_name}')/ListItemAllFields")
        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))

        data = json.loads(response.text)
        return data
    
    @property
    def files(self):        
        response = self._session.get(self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self.folder_name}')/files")
        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))

        data = json.loads(response.text)
        return data['value']
    
    def upload_file(self, content, file_name):
        body = content
        url = self.site_url + f"/_api/web/GetFolderByServerRelativeUrl('{self.folder_name}')/Files/add(url='{file_name}',overwrite=true)"
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}
        
        response = self._session.post(url=url,
                                      headers=headers,
                                      data=body,
                                      timeout=self.timeout)

        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))
        
    def check_out(self, file_name): 
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{file_name}')/CheckOut()"
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = self._session.post(url=url,
                                      headers=headers)
        
        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))

        return None
    
    def check_in(self, file_name, comment):
        url = self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{file_name}')/CheckIn(comment='{comment}',checkintype=0)"
    
        headers = {'X-RequestDigest': self.contextinfo['FormDigestValue']}

        response = self._session.post(url=url,
                                      headers=headers)

        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))
        
        return None
    
    def read_txt_file(self, file_name):
        response = self._session.get(self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{file_name}')/$value")
        if response.status_code != 200:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))
        return response.text

    
    def download_file(self, source_file_name, dest_file_name):
        """
            downloads source_file_name from the sharepiont folder and saves it to dest_file_name
        """
        response = self._session.get(self.site_url + f"/_api/web/GetFileByServerRelativeUrl('{self.info['d']['ServerRelativeUrl']}/{source_file_name}')/$value")
        if response.status_code == 200:
            with open (dest_file_name, 'wb') as f:
                f.write(response.content)
        else:
            response.raise_for_status()
            raise RuntimeError("Response code: " + str(response.status_code) + ", response: " + str(response.text))


# SharePlum
# This library simplfies the code necessary
# to automate interactions with a SharePoint
# server using python
from .office365 import Office365  # noqa: F401
from .site import Site  # noqa: F401
from .version import __version__  # noqa: F401

__all__ = ["site", "office365"]

__title__ = "SharePlum SharePoint Library"
__author__ = "Jason Rollins"
